/*******************************************************************@CM2CPYR*/
/*                                                                  @CM2CPYR*/
/*     Licensed Materials - Property of IBM                         @CM2CPYR*/
/*                                                                  @CM2CPYR*/
/*     5622-078                                                     @CM2CPYR*/
/*     (C) Copyright IBM Corp. 1988, 1995 All Rights Reserved.      @CM2CPYR*/
/*                                                                  @CM2CPYR*/
/*     US Government Users Restricted Rights - Use, duplication or  @CM2CPYR*/
/*     disclosure restricted by GSA ADP Schedule Contract with      @CM2CPYR*/
/*     IBM Corp.                                                    @CM2CPYR*/
/*                                                                  @CM2CPYR*/
/*******************************************************************@CM2CPYR*/
/*****************************************************************************/
/*                                                                           */
/*   Component   : Port Connection Manager (PCM)                             */
/*                                                                           */
/*   Module      : PCMINIT.C                                                 */
/*                                                                           */
/*   Description : This module contains the routines for starting and        */
/*                 stopping the PCM.  This module also contains the          */
/*                 GCCI routines for activating and deactivating SAPs.       */
/*                                                                           */
/*****************************************************************************/

#define INCL_DOS
#include <os2.h>
#include <memory.h>
#include "gcci_com.h"
#include "gcci_cm.h"
#include "sf.h"
#include "cci_rc.h"
#include "pcmscpi.h"
#include "pcmdef.h"
#include "pcmproto.h"


extern HSYSSEM QueueSem;    /* Queue Semaphore */
extern HSYSSEM TimerSem;    /* Timer Semaphore */

extern PPPRIM pQueue;       /* Pointer to linked list of primitives */
extern PTIMER TimerQueue;   /* Linked list of timer sturctures */

extern SEL  QueueStackSel;  /* Selector for ProcessQueue stack segment  */
extern SEL  TimerStackSel;  /* Selector for ProcessTimers stack segment */

extern BOOL StopTimers;     /* Set TRUE when ProcessTimers is being stopped */
extern BOOL StopQueue;      /* Set TRUE when ProcessQueue is being stopped  */
extern ULONG TimersStoppedSem; /* Cleared by ProcessTimers when it has stopped */
extern ULONG QueueStoppedSem;  /* Cleared by ProcessQueue when it has stopped  */

USHORT PCMHandle;    /* Handle for PCM device driver */

PPCMCB     pPCMCBList = NULL;  /* Linked list of PCM control blocks */

BOOL DLLStarted = FALSE;


/*****************************************************************************/
/*                                                                           */
/*   Routine     : StartVPCM                                                 */
/*                                                                           */
/*   Description : This routine is called by the Module Enabler to start     */
/*                 one or more PCMs.                                         */
/*                                                                           */
/*                 If no PCMs are currently started, then the DLL is         */
/*                 started first.  The requested PCMs are then started.      */
/*                 If no PCM was successfully started, then the DLL is       */
/*                 shutdown.                                                 */
/*                                                                           */
/*****************************************************************************/

USHORT _export _loadds pascal StartVPCM( GCCI_COM_MER_INITPCM *pInitCB )
{
   USHORT RetCode;

      return RetCode;
      return CCI_LACK_OF_RESOURCES;
      return CCI_NO_ERROR;
      /* etc */
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ExitDLL                                                   */
/*                                                                           */
/*   Description : This is the exit list routine that is executed if the     */
/*                 Moduler Enabler process is abnormally stopped.  This      */
/*                 routine shuts down all active PCMs and releases all       */
/*                 global resources.  The PCM device driver is closed.       */
/*                                                                           */
/*****************************************************************************/

VOID _loadds pascal far ExitDLL( USHORT ExitType )
{
      switch ( ExitType )
         {
            case TC_EXIT:        printf("Normal Exit\n");    break;
            case TC_HARDERROR:   printf("Hard Error\n");     break;
            case TC_TRAP:        printf("Trap\n");           break;
            case TC_KILLPROCESS: printf("DosKillProcess\n"); break;
            case TC_EXCEPTION:   printf("Exception\n");      break;
         }


   /* Close the PCM device driver */
   DosClose( PCMHandle );

   /* Exit this routine */
   DosExitList( EXLST_EXIT, ExitDLL );
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : StartupPCM                                                */
/*                                                                           */
/*   Description : This routine starts a new PCM.  If the requested PCM      */
/*                 is already started, then the PCM is not started again.    */
/*                                                                           */
/*                 The tasks performed by this routine are:                  */
/*                   - Get configuration data from Module Enabler            */
/*                   - Create a new PCM control block                        */
/*                   - Process the PCM's configuration data                  */
/*                   - Startup the ports for this PCM                        */
/*                   - Compile the PCM's Dialog Descriptor File              */
/*                   - Register the PCM to the Module Enabler                */
/*                                                                           */
/*                 This routine does not pass back a return code.            */
/*                 If the PCM is successfully started, then the PCM is       */
/*                 registered with the Moduler Enabler and a PCMCB is        */
/*                 added to the list of PCMCBs.  If a PCM can not be         */
/*                 started, then an error message is generated and the       */
/*                 PCM is not registered with the Module Enabler.            */
/*                                                                           */
/*****************************************************************************/

VOID near StartupPCM( PUCHAR pPCMName, GCCI_COM_MER_INITPCM *pInitCB )
{
   GCCI_COM_CI_GET_CONFIG GetConfig;
   GCCI_COM_CI_REG_MODULE RegModule;
   BOOL Startup;
   USHORT RetCode;

   /* Put PCM Name in GetConfig control information */
   if ( !memcpy( GetConfig.PCMName, pPCMName, 16 ) )
      {
         /* Setup GetConfig control information */
         GetConfig.CtrlInfoLength = sizeof(GCCI_COM_CI_GET_CONFIG);
         GetConfig.PrimitiveCode  = GCCI_COM_PC_GET_CONFIG;
         GetConfig.DataInfoLength = 0;
         GetConfig.Selector       = 0;
         GetConfig.IDType         = GCCI_COM_ID_TYPE_NONE;
         GetConfig.Identifier     = 0L;

         /* Call Module Enabler to get config data */
         RetCode = pInitCB->MERRegEP(&GetConfig);
         if ( RetCode == CCI_NO_ERROR )
            {
               /* Process configuration data */
                  /* Startup ports */
                  /* Put PCM Name in RegModule control information */
                  if ( !memcpy( RegModule.PCMName, pPCMName, 16 ) )
                     {
                        /* Setup RegModule control information */
                        RegModule.PrimitiveCode  = GCCI_COM_PC_REG_MODULE;
                        RegModule.DataInfoLength = 0;
                        RegModule.Selector       = 0;
                        RegModule.IDType         = GCCI_COM_ID_TYPE_NONE;
                        RegModule.Identifier     = 0L;
                        RegModule.PCMActSapEP    = EnqueueGCCI;
                        RegModule.PCM_Version    = 0;

                        /* Call Module Enabler to register PCM */
                        if ( pInitCB->MERRegEP( &RegModule ) )
                           {
                              Startup = FAILURE;
                            }
                     }
            }
       }
}
/*****************************************************************************/
/*                                                                           */
/*   Routine     : ShutdownPCM                                               */
/*                                                                           */
/*   Description : This routine stops a PCM.  The PCM's control block,       */
/*                 dialog descriptor table, SAPs and call instances are      */
/*                 deleted.  The ports for this PCM are shutdown.            */
/*                                                                           */
/*****************************************************************************/

VOID near ShutdownPCM( PUCHAR pPCMName )
{
   PPCMCB pPCMCB;


   /* If CM SAP has not been deactivated */
         /* Delete all active Call Instances */

         /* CM SAP is now deactivated */

   ShutdownPorts;

   pPCMCB->Status = STATUS_DEACTIVATED;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ActivateSAP                                               */
/*                                                                           */
/*   Description : This routine processes an Activate SAP request primitive. */
/*                                                                           */
/*                 An Activate SAP request primitive will be received        */
/*                 from the Connection Manager and another from the CNM      */
/*                 manager for each PCM that was registered to the           */
/*                 Module Enabler.                                           */
/*                                                                           */
/*                 This routine creates a SAP and attaches it to the         */
/*                 correct PCM control block.  The entry points to the       */
/*                 Connection Manager or the CNM manager are stored in       */
/*                 the SAP.                                                  */
/*                                                                           */
/*                 If the Activate SAP request primitive came from the       */
/*                 CNM manager then this routine sends an Activate SAP       */
/*                 confirm primitive back to the CNM manager.                */
/*                                                                           */
/*                 If the Activate SAP request primitive came from the       */
/*                 Connection Manager, then a bogus call instance is         */
/*                 created for each port of the PCM.  For each port, an      */
/*                 instance of the state machine is spawned to activate      */
/*                 the port.  The state machine instances are collected      */
/*                 back together at the PortDone routine.                    */
/*                                                                           */
/*****************************************************************************/

VOID near ActivateSAP( GCCI_COM_CI_ACT_SAP_REQ *pRequest )
{
   GCCI_COM_SF01_ACT_SAP_REQ *pLowerDsp;
   PPCMCB  pPCMCB;
   PPORT   pPort;
   BOOL    PortsAvailable;

   /* Find the PCM control block for this SAP */
   for ( pPCMCB = pPCMCBList;
         !memcmp( pRequest->PCMName, pPCMCB->PCMName, 16 );
         pPCMCB = pPCMCB->pNextPCMCB )
      if ( pPCMCB->pNextPCMCB == NULL )
         {
            return;
         }

   SELECTOROF(pLowerDsp) = SELECTOROF(pRequest);
   OFFSETOF(pLowerDsp)   = OFFSETOF(pRequest)+sizeof(GCCI_COM_CI_ACT_SAP_REQ);

   pPCMCB->Status = STATUS_ACTIVATING;

   if ( pRequest->ServiceType & GCCI_COM_SRV_TYPE_CNM_MGR )  /* CNM Manager */
      {
         pPCMCB->CNMSAP.Status    = STATUS_ACTIVATED;
         pPCMCB->CNMSAP.U_SAP_ID  = pRequest->U_SAP_ID;
         pPCMCB->CNMSAP.Indicate  = pLowerDsp->SFCNMIndEP;
         pPCMCB->CNMSAP.Confirm   = pLowerDsp->SFCNMCnfEP;

         CNMActivateSAPConfirm( pPCMCB );

         /* If both SAPs are now activated then the PCM is activated */
         if ( pPCMCB->CMSAP.Status == STATUS_ACTIVATED )
            pPCMCB->Status = STATUS_ACTIVATED;
      }
   else if ( pRequest->ServiceType & GCCI_COM_SRV_TYPE_CONNECT_MGR )  /* CM */
      {
         pPCMCB->CMSAP.U_SAP_ID = pRequest->U_SAP_ID;
         pPCMCB->CMSAP.Indicate = pLowerDsp->SFCMIndEP;
         pPCMCB->CMSAP.Confirm  = pLowerDsp->SFCMCnfEP;

         /* Check if there are ports available to be activated */
         for ( pPort=pPCMCB->pPortList,PortsAvailable=FALSE; pPort!=NULL;
               pPort=pPort->pNextPort )
            {
               if ( pPort->Status == STATUS_ACTIVATING )
                  {
                     PortsAvailable = TRUE;
                     break;
                  }
            }

         if ( PortsAvailable )   /* If there are ports available */
            {
               pPCMCB->CMSAP.Status = STATUS_ACTIVATING;

               for ( pPort=pPCMCB->pPortList; pPort!=NULL;
                     pPort=pPort->pNextPort )
                  {
                     if ( pPort->Status == STATUS_ACTIVATING )
                        ActivatePort( pPort );
                  }
            }
         else  /* No ports are available */
            {
               pPCMCB->CMSAP.Status = STATUS_DEACTIVATED;

               /* Send unsuccessful Activate SAP Confirm */
               CMActivateSAPConfirm( pPCMCB, CCI_RESOURCE_NOT_AVAILABLE );

               /* Deactivate CNM SAP if activated */
               if ( pPCMCB->CNMSAP.Status == STATUS_ACTIVATED )
                  {
                     DeactivateSAPIndicate ( &(pPCMCB->CNMSAP), CCI_RESOURCE_NOT_AVAILABLE );

                     /* Stop all traces */
                     pPCMCB->Flags.GCCITraceActive   = FALSE;
                     pPCMCB->Flags.SCPITraceActive   = FALSE;
                     pPCMCB->Flags.EventTraceActive  = FALSE;
                     pPCMCB->Flags.DialogTraceActive = FALSE;

                     /* SAP is now deactivated */
                     pPCMCB->CNMSAP.Status = STATUS_DEACTIVATED;

                     /* Shutdown PCM */
                     ShutdownPCM( pPCMCB->PCMName );
                  }
            }
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CreatePort                                                */
/*                                                                           */
/*   Description : This routine creates a new Port structure,                */
/*                 initializes it, and adds it to the end of the list        */
/*                 of Port.                                                  */
/*                                                                           */
/*****************************************************************************/

PPORT near CreatePort( PPCMCB pPCMCB )
{
   PPORT pPort, pLastPort;

   /* Create a new Port */

   /* Initialize the Port variables */
   pPort->pNextPort   = NULL;
   pPort->Status      = STATUS_ACTIVATING;
   pPort->Phase       = PHASE_ACTIVATING;
   pPort->State       = 0;
   pPort->Name[0]     = 0;
   pPort->LocalNum[0] = 0;
   pPort->pPPAT       = NULL;
   pPort->pPCMCB      = pPCMCB;

   /* Initialize Port flags */
   pPort->Flags.PermConnection = FALSE;
   pPort->Flags.PhysConnActive = FALSE;
   pPort->Flags.DataConnActive = FALSE;
   pPort->Flags.PortInUse      = FALSE;

   /* Add the new Port to the Port list */
   if ( pPCMCB->pPortList == NULL )
      pPCMCB->pPortList = pPort;
   else
      {
         for ( pLastPort=pPCMCB->pPortList;
               pLastPort->pNextPort!=NULL;
               pLastPort=pLastPort->pNextPort );
         pLastPort->pNextPort = pPort;
      }

   return pPort;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeletePort                                                */
/*                                                                           */
/*   Description : This routine removes a Port structure from the            */
/*                 Port list and releases its memory.                        */
/*                                                                           */
/*****************************************************************************/

PPORT near DeletePort( PPORT pPort, PPCMCB pPCMCB )
{
   PPORT pPrevPort, pNextPort;

   /* Remove the Port from the Port list */
   if ( pPCMCB->pPortList == pPort )
      {
         pNextPort = pPCMCB->pPortList = pPort->pNextPort;
      }
   else
      {
         for ( pPrevPort=pPCMCB->pPortList;
               pPrevPort->pNextPort!=pPort;
               pPrevPort=pPrevPort->pNextPort );
         pNextPort = pPrevPort->pNextPort = pPort->pNextPort;
      }

   /* Release memory for the Port structure */

   return pNextPort;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ActivatePort                                              */
/*                                                                           */
/*   Description : This routine is called for each port of a PCM whose       */
/*                 SAP is being activated.  This routine sets a port's       */
/*                 initial PPAT paramters.  After setting the PPAT           */
/*                 parameters, the state machine will enter the              */
/*                 initialization dialog to initialize the port's modem.     */
/*                                                                           */
/*****************************************************************************/

VOID near ActivatePort( PPORT pPort )
{
   PPCMCB pPCMCB;

   pPCMCB = pPort->pPCMCB;

   pPort->Phase  = PHASE_ACTIVATING;
   pPort->Status = STATUS_ACTIVATING;

   if ( GetPPATParms( pPort ) == SCPI_RC_SUCCESS )
      {
         if ( pPCMCB->Flags.Asynchronous )
            {
               pPort->pPPAT->Xmit_Speed = pPCMCB->SerialPortSpeed;
               pPort->pPPAT->Rcv_Speed  = pPCMCB->SerialPortSpeed;

               if ( pPCMCB->Flags.FlowControl )
                  {
                     if ( pPCMCB->Flags.FlowControlType == RTSCTS )
                        pPort->pPPAT->Hard_Hand = 0x0C;  /* RTS/CTS  */
                     else
                        pPort->pPPAT->Hard_Hand = 0x03;  /* XON/XOFF */
                  }
               else pPort->pPPAT->Hard_Hand = 0x00;      /* No Flow Control */

               pPort->pPPAT->Port_Trans_Mode = 0x0008;  /* 8-N-1 */
            }
         else  /* Synchronous Port */
            {
               pPort->pPPAT->Port_Trans_Mode = 0x00;  /* NRZ */
            }

         pPort->pPPAT->Signal_Mask = 0x0000;  /* No Signals */
         pPort->pPPAT->Hw_Err_Mask = 0xFFFF;  /* All Errors */

         /* Set PPAT */
      }
   else  /* Can not get PPAT parameters */
      {
         DeactivatePort( pPort );
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeactivatePort                                            */
/*                                                                           */
/*   Description : This routine deactivates a port by setting the            */
/*                 PortActivated flag to FALSE and deactivating the          */
/*                 port's physical connection.                               */
/*                                                                           */
/*****************************************************************************/

VOID near DeactivatePort( PPORT pPort )
{
   pPort->Phase  = PHASE_DEACTIVATING;
   pPort->Status = STATUS_DEACTIVATING;

   /* Deactivate Connection */
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : PortDone                                                  */
/*                                                                           */
/*   Description : This routine is called when a port's activation or        */
/*                 deactivation is complete.  The port's bogus call          */
/*                 instance is deleted.                                      */
/*                                                                           */
/*                 If a SAP is being activated and the last port has         */
/*                 completed its activation attempt, then delete all the     */
/*                 ports that failed to activate.  If at least one port      */
/*                 was successfully activated, then send a successfull       */
/*                 Activate SAP confirm primitive to the Connection          */
/*                 Manager.  If all ports failed to activate, then send      */
/*                 and Activate SAP confirm primitive indicating that        */
/*                 the SAP could not be activated and delete the SAP.        */
/*                                                                           */
/*                 If a SAP is being deactivated and the last port has       */
/*                 completed its deactivation, then send a Deactivate SAP    */
/*                 confirm primitive to the Connection Manager and delete    */
/*                 the SAP.                                                  */
/*                                                                           */
/*****************************************************************************/

VOID near PortDone( PPORT pPortDone )
{
   PPCMCB pPCMCB;
   PPORT  pPort;
   USHORT PortCnt;

   pPCMCB = pPortDone->pPCMCB;

   /* Make sure all timers for this port are stopped */
   StopTimer( TIMER_CLASS_STATE_TIMER, pPort );

   if ( pPortDone->Status == STATUS_ACTIVATING )
      {
         pPortDone->Status = STATUS_ACTIVATED;
      }
   else if ( pPortDone->Status == STATUS_DEACTIVATING )
      {
         pPortDone->Status = STATUS_DEACTIVATED;

         /* Give up ownwership of the port */

         /* Release the PPAT buffer */

      }

   /* Check if all ports are done */
   for (pPort=pPCMCB->pPortList,PortCnt=0;pPort!=NULL;pPort=pPort->pNextPort)
      if (      pPort->Status == STATUS_ACTIVATING   ) break; /* Not done yet */
      else if ( pPort->Status == STATUS_DEACTIVATING ) break; /* Not done yet */
      else if ( pPort->Status == STATUS_ACTIVATED    ) PortCnt++;

   if ( pPort == NULL )  /* If all ports are done */
      {
         if ( pPCMCB->CMSAP.Status == STATUS_ACTIVATING )
            {
               if ( PortCnt > 0 )  /* If at least one port was activated */
                  {
                     /* SAP is activated */
                     pPCMCB->CMSAP.Status = STATUS_ACTIVATED;

                     /* Send successful Activate SAP Confirm */
                     CMActivateSAPConfirm( pPCMCB, CCI_NO_ERROR );

                     /* If both SAPs are now activated then PCM is activated */
                     if ( pPCMCB->CNMSAP.Status == STATUS_ACTIVATED )
                        pPCMCB->Status = STATUS_ACTIVATED;

                     /* Make all activated ports idle */
                     for ( pPort=pPCMCB->pPortList; pPort!=NULL;
                           pPort=pPort->pNextPort)
                        {
                           if ( pPort->Status == STATUS_ACTIVATED )
                              PortIdle( pPort );
                        }
                  }
               else  /* No ports were activated */
                  {
                     /* SAP could not be activated */
                     pPCMCB->CMSAP.Status = STATUS_DEACTIVATED;

                     /* Send unsuccessful Activate SAP Confirm */
                     CMActivateSAPConfirm( pPCMCB, CCI_RESOURCE_NOT_AVAILABLE );


                     /* Deactivate CNM SAP if activated */
                     if ( pPCMCB->CNMSAP.Status == STATUS_ACTIVATED )
                        {
                           DeactivateSAPIndicate ( &(pPCMCB->CNMSAP), CCI_RESOURCE_NOT_AVAILABLE );

                           /* Stop all traces */
                           pPCMCB->Flags.GCCITraceActive   = FALSE;
                           pPCMCB->Flags.SCPITraceActive   = FALSE;
                           pPCMCB->Flags.EventTraceActive  = FALSE;
                           pPCMCB->Flags.DialogTraceActive = FALSE;

                           /* SAP is now deactivated */
                           pPCMCB->CNMSAP.Status = STATUS_DEACTIVATED;

                           /* Shutdown PCM */
                           ShutdownPCM( pPCMCB->PCMName );
                        }

                  }
            }
         else if ( pPCMCB->CMSAP.Status == STATUS_DEACTIVATING )
            {
               /* SAP is now deactivated */
               pPCMCB->CMSAP.Status = STATUS_DEACTIVATED;

               /* Send Deactivate SAP Confirm */
               DeactivateSAPConfirm( &(pPCMCB->CMSAP) );

               /* If both SAPs are now deactivated then shutdown the PCM */
               if ( pPCMCB->CNMSAP.Status == STATUS_DEACTIVATED )
                 ShutdownPCM( pPCMCB->PCMName );
            }
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : PortIdle                                                  */
/*                                                                           */
/*   Description :                                                           */
/*                                                                           */
/*****************************************************************************/

VOID near PortIdle( PPORT pPort )
{
   pPort->Phase = PHASE_IDLE;

   if ( pPort->Flags.PermConnection )
      {
         /* Event Trace of the state */

         pPort->State = STATE_IDLE;
      }
   else  /* Port is not a permanent connection */
      {
         /* Enter the "RING" dialog and wait for the ring string to arrive */
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CNMActivateSAPConfirm                                     */
/*                                                                           */
/*   Description : This routine sends an Activate SAP confirm primitive      */
/*                 to the CNM manager.                                       */
/*                                                                           */
/*                 This primitive includes subfields that define the         */
/*                 PCM's trace points and dump to the CNM manager.           */
/*                                                                           */
/*****************************************************************************/

VOID near CNMActivateSAPConfirm( PPCMCB pPCMCB )
{
   GCCI_COM_CI_ACT_SAP_CNF   *pConfirm;
   GCCI_COM_SF02_ACT_SAP_CNF *pUpperDsp;
   GCCI_COM_SF03_ACT_SAP_CNF *pCNMInfo;
   GCCI_COM_TRACE_PT_DEF     *pTracePt;
   GCCI_COM_DUMP_DEF         *pDump;
   USHORT ConfirmSize, CNMInfoSize;
   USHORT RetCode, x;

   /* Calculate size of Activate SAP Confirm primitive */
   CNMInfoSize = sizeof(GCCI_COM_SF03_ACT_SAP_CNF)   +
                 (sizeof(GCCI_COM_TRACE_PT_DEF) * 4) +
                 (sizeof(GCCI_COM_DUMP_DEF)     * 1);
   ConfirmSize = sizeof(GCCI_COM_CI_ACT_SAP_CNF)   +
                 sizeof(GCCI_COM_SF02_ACT_SAP_CNF) +
                 CNMInfoSize;

   /* Get memory for Activate SAP Confirm primitive */

   /* Fill in Activate SAP Confirm Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(GCCI_COM_CI_ACT_SAP_CNF);
   pConfirm->PrimitiveCode  = GCCI_COM_PC_ACT_SAP_CNF;
   pConfirm->DataInfoLength = CNMInfoSize + sizeof(GCCI_COM_SF02_ACT_SAP_CNF);
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pConfirm->U_SAP_ID       = pPCMCB->CNMSAP.U_SAP_ID;
   pConfirm->P_SAP_ID       = (DWORD)(&(pPCMCB->CNMSAP));
   for ( x=0; x<16; x++ ) pConfirm->Reserved1[x] = 0;
   pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.ReportCompCode   = CCI_NO_ERROR;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.OrigCompCode     = CCI_NO_ERROR;
   pConfirm->CommonStatus.SysLogID         = 0L;

   /* Fill in CNM Info Subfield */
   SELECTOROF(pCNMInfo)     = SELECTOROF(pConfirm);
   OFFSETOF(pCNMInfo)       = OFFSETOF(pConfirm) + sizeof(GCCI_COM_CI_ACT_SAP_CNF);
   pCNMInfo->SFLength       = CNMInfoSize;
   pCNMInfo->SFIdentifier   = GCCI_COM_SFID_PRM_SPECIFIC;
   pCNMInfo->SFType         = GCCI_COM_SF_CNM_DATA;
   pCNMInfo->NumTracePts    = 4;
   pCNMInfo->SizeofTracePts = sizeof(GCCI_COM_TRACE_PT_DEF);
   pCNMInfo->NumDumps       = 1;
   pCNMInfo->SizeofDumps    = sizeof(GCCI_COM_DUMP_DEF);

   /* Fill in Trace Point for GCCI API trace */
   SELECTOROF(pTracePt) = SELECTOROF(pConfirm);
   OFFSETOF(pTracePt)   = OFFSETOF(pCNMInfo) + sizeof(GCCI_COM_SF03_ACT_SAP_CNF);
   pTracePt->TracePtID  = TRACE_ID_GCCI_API;
   pTracePt->TraceType  = GCCI_CNM_TRACE_TYPE_API;
   memcpy(pTracePt->TraceName, "GCCI", GCCI_COM_MAX_TRACENAME_LEN);

   /* Fill in Trace Point for SCPI API trace */
   OFFSETOF(pTracePt)  += sizeof(GCCI_COM_TRACE_PT_DEF);
   pTracePt->TracePtID  = TRACE_ID_SCPI_API;
   pTracePt->TraceType  = GCCI_CNM_TRACE_TYPE_API;
   memcpy( pTracePt->TraceName, "SCPI", GCCI_COM_MAX_TRACENAME_LEN);

   /* Fill in Trace Point for event trace */
   OFFSETOF(pTracePt)  += sizeof(GCCI_COM_TRACE_PT_DEF);
   pTracePt->TracePtID  = TRACE_ID_EVENT;
   pTracePt->TraceType  = GCCI_CNM_TRACE_TYPE_EVENT;
   memcpy( pTracePt->TraceName, "EVENT", GCCI_COM_MAX_TRACENAME_LEN);

   /* Fill in Trace Point for dialog trace */
   OFFSETOF(pTracePt)  += sizeof(GCCI_COM_TRACE_PT_DEF);
   pTracePt->TracePtID  = TRACE_ID_DIALOG;
   pTracePt->TraceType  = GCCI_CNM_TRACE_TYPE_EVENT;
   memcpy( pTracePt->TraceName, "DIALOG", GCCI_COM_MAX_TRACENAME_LEN);

   /* Fill in Dump definition */
   SELECTOROF(pDump) = SELECTOROF(pConfirm);
   OFFSETOF(pDump)   = OFFSETOF(pTracePt) + sizeof(GCCI_COM_TRACE_PT_DEF);
   pDump->DumpID     = 1;
   pDump->Reserved   = 0;
   memcpy( pDump->DumpName, "CTRL BLOCK", GCCI_COM_MAX_TRACENAME_LEN);

   /* Fill in Upper Dispatch Subfield */
   SELECTOROF(pUpperDsp)   = SELECTOROF(pConfirm);
   OFFSETOF(pUpperDsp)     = OFFSETOF(pCNMInfo) + CNMInfoSize;
   pUpperDsp->SFLength     = sizeof(GCCI_COM_SF02_ACT_SAP_CNF);
   pUpperDsp->SFIdentifier = GCCI_COM_SFID_PRM_SPECIFIC;
   pUpperDsp->SFType       = GCCI_COM_SF_UPPER_DISPATCH;
   pUpperDsp->SFCCReqEP    = EnqueueGCCI;
   pUpperDsp->SFCCRspEP    = EnqueueGCCI;
   pUpperDsp->SFRASReqEP   = EnqueueGCCI;
   pUpperDsp->SFRASRspEP   = EnqueueGCCI;

   /* Call CNM Manager with Activate SAP Confirm primitive */
   RetCode = pPCMCB->CNMSAP.Confirm( pConfirm );

   /* Release memory allocated for Activate SAP Control Block */
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CMActivateSAPConfirm                                      */
/*                                                                           */
/*   Description : This routine sends an Activate SAP confrim primitive      */
/*                 with the specified completion code to the Connection      */
/*                 Manager.                                                  */
/*                                                                           */
/*                 This primitive includes subfields that informs the        */
/*                 Connection Manager which ports and local phone numbers    */
/*                 this PCM manages.                                         */
/*                                                                           */
/*****************************************************************************/

VOID near CMActivateSAPConfirm( PPCMCB pPCMCB, USHORT CompCode )
{
   GCCI_COM_CI_ACT_SAP_CNF    *pConfirm;
   GCCI_COM_SF01X_ACT_SAP_CNF *pPCMInfo;
   GCCI_COM_SF02_ACT_SAP_CNF  *pUpperDsp;
   USHORT ConfirmSize, PCMInfoSize;
   USHORT RetCode, x;

   /* Calculate size of Activate SAP Confirm primitive */
   PCMInfoSize = sizeof( GCCI_COM_SF01X_ACT_SAP_CNF );

   ConfirmSize = sizeof( GCCI_COM_CI_ACT_SAP_CNF) +
                 sizeof( GCCI_COM_SF02_ACT_SAP_CNF) +
                 sizeof( GCCI_COM_SF01X_ACT_SAP_CNF );

   /* Get memory for Activate SAP Confirm primitive */

   /* Fill in Activate SAP Confirm Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(GCCI_COM_CI_ACT_SAP_CNF);
   pConfirm->PrimitiveCode  = GCCI_COM_PC_ACT_SAP_CNF;
   pConfirm->DataInfoLength = PCMInfoSize + sizeof(GCCI_COM_SF02_ACT_SAP_CNF);
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pConfirm->U_SAP_ID       = pPCMCB->CMSAP.U_SAP_ID;
   pConfirm->P_SAP_ID       = (DWORD)(&(pPCMCB->CMSAP));
   for ( x=0; x<16; x++ ) pConfirm->Reserved1[x] = 0;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.ReportCompCode   = CompCode;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.OrigCompCode     = CompCode;
   pConfirm->CommonStatus.SysLogID         = 0L;

   if ( CompCode == CCI_NO_ERROR )
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   else  /* CompCode is an error */
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;

   /* Fill in PCM Info Subfield */
   SELECTOROF(pPCMInfo)    = SELECTOROF(pConfirm);
   OFFSETOF(pPCMInfo)      = OFFSETOF(pConfirm) + sizeof(GCCI_COM_CI_ACT_SAP_CNF);
   pPCMInfo->SFLength      = PCMInfoSize;
   pPCMInfo->SFIdentifier  = GCCI_COM_SFID_PRM_SPECIFIC;
   pPCMInfo->SFType        = GCCI_COM_SF_PCM_INFORMATION;
   pPCMInfo->NetworkType   = GCCI_COM_NET_TYPE_X25;          /* NETWORK TYPE */

   /* Fill in Channel Information */
   pPCMInfo->ChanInfo.XMacAddr = GetMacCCT();

   pPCMInfo->LocalNum.NumberType = SF_NUMBER_TYPE_UNSPECIFIED;
   pPCMInfo->LocalNum.NumberPlan = SF_NUMBER_PLAN_UNKNOWN;
   pPCMInfo->LocalNum.NumberLength = 0;


   /* Fill in Upper Dispatch Subfield */
   SELECTOROF(pUpperDsp)   = SELECTOROF(pConfirm);
   OFFSETOF(pUpperDsp)     = OFFSETOF(pPCMInfo) + PCMInfoSize;
   pUpperDsp->SFLength     = sizeof(GCCI_COM_SF02_ACT_SAP_CNF);
   pUpperDsp->SFIdentifier = GCCI_COM_SFID_PRM_SPECIFIC;
   pUpperDsp->SFType       = GCCI_COM_SF_UPPER_DISPATCH;
   pUpperDsp->SFCCReqEP    = EnqueueGCCI;
   pUpperDsp->SFCCRspEP    = EnqueueGCCI;
   pUpperDsp->SFRASReqEP   = EnqueueGCCI;
   pUpperDsp->SFRASRspEP   = EnqueueGCCI;

   /* Call Connection Manager with Activate SAP Confirm primitive */
   RetCode = pPCMCB->CMSAP.Confirm( pConfirm );

   /* Free memory allocated for Activate SAP Control Block */
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeactivateSAP                                             */
/*                                                                           */
/*   Description : This routine processes a Deactivate SAP request           */
/*                 primitive.                                                */
/*                                                                           */
/*                 If the Deactivate SAP request primitive came from the     */
/*                 CNM manager then this routine sends a Deactivate SAP      */
/*                 confirm primitive back to the CNM manager and deletes     */
/*                 the SAP.                                                  */
/*                                                                           */
/*                 If the Deactivate SAP request primitive came from the     */
/*                 Connection Manager, then a bogus call instance is         */
/*                 created for each port of the PCM.  For each port, an      */
/*                 instance of the state machine is spawned to deactivate    */
/*                 the port.  The state machine instances are collected      */
/*                 back together at the PortDone routine.                    */
/*                                                                           */
/*****************************************************************************/

VOID near DeactivateSAP( GCCI_COM_CI_DEACT_SAP_REQ *pRequest )
{
   PPCMCB pPCMCB;
   PSAP   pSAP;
   PPORT  pPort;

   pSAP   = (PSAP)pRequest->P_SAP_ID;
   pPCMCB = pSAP->pPCMCB;

   /* Stop all traces */
   pPCMCB->Flags.GCCITraceActive   = FALSE;
   pPCMCB->Flags.SCPITraceActive   = FALSE;
   pPCMCB->Flags.EventTraceActive  = FALSE;
   pPCMCB->Flags.DialogTraceActive = FALSE;

   /* This PCM is now deactivating */
   pPCMCB->Status = STATUS_DEACTIVATING;

   if ( pSAP == &(pPCMCB->CNMSAP) )     /* Deactivating CNM SAP */
      {
         /* SAP is now deactivated */
         pPCMCB->CNMSAP.Status = STATUS_DEACTIVATED;

         /* Send Deactivate SAP Confirm */
         DeactivateSAPConfirm( pSAP );

         /* If both SAPs are now deactivated then shutdown the PCM */
         if ( pPCMCB->CMSAP.Status == STATUS_DEACTIVATED )
            ShutdownPCM( pPCMCB->PCMName );
      }
   else if ( pSAP == &(pPCMCB->CMSAP) )  /* Deactivating CM SAP */
      {
         /* Delete all active Call Instances */
         for ( pPort=pPCMCB->pPortList; pPort!=NULL; pPort=pPort->pNextPort )
            {
               if ( pPort->Flags.PortInUse )
                  {
                     /* Ensure the MAC is in Connection mode */

                     /* Active calls will be terminated by DeactivatePort() */
                  }
            }

         /* This SAP is now deactivating */
         pSAP->Status = STATUS_DEACTIVATING;

         /* Mark all activated ports as deactivating */
         for ( pPort=pPCMCB->pPortList; pPort!=NULL; pPort=pPort->pNextPort )
            {
               if ( pPort->Status == STATUS_ACTIVATED )
                  pPort->Status = STATUS_DEACTIVATING;
            }

         /* Deactivate all activated ports */
         for ( pPort=pPCMCB->pPortList; pPort!=NULL; pPort=pPort->pNextPort )
            {
               if ( pPort->Status == STATUS_DEACTIVATING )
                  DeactivatePort( pPort );
            }
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeactivateSAPConfirm                                      */
/*                                                                           */
/*   Description : This routine sends a Deactivate SAP confirm primitive     */
/*                 to the Connection Manager or the CNM manager.             */
/*                                                                           */
/*****************************************************************************/

VOID near DeactivateSAPConfirm( PSAP pSAP )
{
   GCCI_COM_CI_DEACT_SAP_CNF *pConfirm;
   USHORT RetCode;

   /* Get memory for Deactivate SAP Confirm primitive */

   /* Fill in Deactivate SAP Confirm Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(GCCI_COM_CI_DEACT_SAP_CNF);
   pConfirm->PrimitiveCode  = GCCI_COM_PC_DEACT_SAP_CNF;
   pConfirm->DataInfoLength = 0;
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pConfirm->U_SAP_ID       = pSAP->U_SAP_ID;
   pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.ReportCompCode   = CCI_NO_ERROR;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.OrigCompCode     = CCI_NO_ERROR;
   pConfirm->CommonStatus.SysLogID         = 0L;

   /* Call CM or CNM with Deactivate SAP Confirm primitive */
   RetCode = pSAP->Confirm( pConfirm );

   /* Free memory for Deactivate SAP Confirm primitive */
}



/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeactivateSAPIndicate                                     */
/*                                                                           */
/*   Description : This routine sends a Deactivate SAP indicate primitive    */
/*                 to the CM or CNM Manager.                                 */
/*                                                                           */
/*****************************************************************************/

BOOL near DeactivateSAPIndicate( PSAP pSAP, USHORT CompCode )
{
   GCCI_COM_CI_DEACT_SAP_IND *pIndicate;
   USHORT RetCode;

   /* Get memory for Deactivate SAP Indicate primitive */

   /* Fill in Call Clear Indicate Control Information Field */
   pIndicate->CtrlInfoLength = sizeof(GCCI_COM_CI_DEACT_SAP_IND);
   pIndicate->PrimitiveCode  = GCCI_COM_PC_DEACT_SAP_IND;
   pIndicate->DataInfoLength = 0;
   pIndicate->Selector       = GCCI_COM_SEL_PCM;
   pIndicate->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pIndicate->U_SAP_ID       = pSAP->U_SAP_ID;
   pIndicate->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.ReportCompCode   = CompCode;
   pIndicate->CommonStatus.Reserved         = 0;
   pIndicate->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.OrigCompCode     = CompCode;
   pIndicate->CommonStatus.SysLogID         = 0L;

   if ( CompCode == CCI_NO_ERROR )
      pIndicate->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   else  /* CompCode is an error */
      pIndicate->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;

   /* Call CM or CNM Manager with Deactivate SAP Indicate primitive */
   RetCode = pSAP->Indicate( pIndicate );

   if ( RetCode == CCI_NO_ERROR )
      {
         /* Free memory for Deactivate SAP Indicate primitive */

         return SUCCESS;
      }
   else  /* Error returned on Deactivate SAP Indicate */
      {
         /* Free memory for Deactivate SAP Indicate primitive */

         return FAILURE;
      }
}



/*****************************************************************************/
/*                                                                           */
/*   Routine     : GetPPATParms                                              */
/*                                                                           */
/*   Description : This routine calls the MAC to get PPAT.                   */
/*                                                                           */
/*****************************************************************************/

USHORT near GetPPATParms( PPORT pPort )
{

   /* This function get the PPAT from the MAC */

}



/*****************************************************************************/
/*                                                                           */
/*   Routine     : GetMacCCT                                                 */
/*                                                                           */
/*   Description : This routine get Mac CCCT.                                */
/*                                                                           */
/*****************************************************************************/

ULONG near GetMacCCT( VOID )
{

   /* This function get the Mac CCT */

}


